"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const express_1 = require("express");
const hyrest_1 = require("hyrest");
const HTTP = require("http-status-codes");
/**
 * Assembles a list of all routes of all specified controllers with their parameter injections.
 *
 * @param controllerObjects A list of all controllers to assemble the list of routes from.
 *
 * @return An array of all routes present on all supplied controllers.
 */
function listRoutes(controllerObjects) {
    return controllerObjects.reduce((result, controllerObject) => {
        // Fetch all routes from this particular controller and put the parameter injections next to tehm.
        const routes = hyrest_1.getRoutes(controllerObject).map(route => ({
            route,
            queryParameters: hyrest_1.getQueryParameters(controllerObject, route.property),
            bodyParameters: hyrest_1.getBodyParameters(controllerObject, route.property),
            urlParameters: hyrest_1.getUrlParameters(controllerObject, route.property),
            contextParameters: hyrest_1.getContextParameters(controllerObject, route.property),
            controllerObject,
        }));
        result.push(...routes);
        return result;
    }, []);
}
/**
 * A middleware to use with express. Takes a list of controllers as arguments. All controllers will be attached
 * to react and will receive requests.
 *
 * **Example:**
 *
 * ```
 * app.use(resRpc(new ControllerOne(), new ControllerTwo(), ...));
 * ```
 *
 * @param controllerObjects The instances of controllers to pass express's requests to.
 *
 * @return An express router.
 */
function hyrest(...controllerObjects) {
    let contextFactory;
    let defaultAuthorizationMode = hyrest_1.AuthorizationMode.NOAUTH;
    let authorizationCheck;
    // Get the actual `Controller` instances for each @controller decorated object.
    // Throws an error if an instance of a class not decorated with @controller has been passed.
    const controllers = controllerObjects.map(controllerObject => {
        const controller = Reflect.getMetadata("api:controller", controllerObject.constructor);
        if (!controller) {
            const name = controllerObject.constructor.name;
            throw new Error(`Added an object to the Hyrest middleware which is not a @controller. Check ${name}.`);
        }
        return controller;
    });
    // Get a flat list of all routes present on all controllers.
    const routes = listRoutes(controllerObjects);
    const router = express_1.Router();
    routes.forEach(({ route, queryParameters, bodyParameters, urlParameters, contextParameters, controllerObject }) => {
        // Grab the actual method from the instance and the route's property name.
        const routeMethod = route.target[route.property];
        // This handler will be called for every request passing through this middleware.
        const handler = (request, response, next) => __awaiter(this, void 0, void 0, function* () {
            // Check whether the call was authorized.
            const authorizationOptions = hyrest_1.getAuthorization(route.target, route.property);
            const authorizationMode = authorizationOptions ? authorizationOptions.mode : defaultAuthorizationMode;
            const context = typeof contextFactory === "function" ? yield contextFactory(request) : contextFactory;
            if (authorizationMode === hyrest_1.AuthorizationMode.AUTH) {
                if (typeof authorizationCheck !== "function") {
                    next(new Error("Call to an authorized route but no authorization check was provided."));
                    return;
                }
                const authorized = yield authorizationCheck(request, context);
                const extraCheck = typeof (authorizationOptions && authorizationOptions.check) !== "undefined" ?
                    yield authorizationOptions.check(request, context) : true;
                if (!authorized || !extraCheck) {
                    response.status(HTTP.UNAUTHORIZED).send({ message: "Unauthorized." });
                    return;
                }
            }
            // Prepare the arguments to pass into the call to the route based on the parameter inejctions.
            const args = [];
            queryParameters.forEach(({ index, name }) => { args[index] = request.query[name]; });
            bodyParameters.forEach(({ index }) => { args[index] = request.body; });
            urlParameters.forEach(({ index, name }) => { args[index] = request.params[name]; });
            contextParameters.forEach(({ index }) => { args[index] = context; });
            let data;
            // Validate and convert all values,
            const processed = [];
            const queryErrors = {};
            let bodyError;
            const paramErrors = {};
            let errorEncountered = false;
            yield Promise.all(args.map((arg, index) => __awaiter(this, void 0, void 0, function* () {
                const { fullValidator } = hyrest_1.getParameterValidation(route.target, route.property, index);
                const validationResult = fullValidator ?
                    yield fullValidator(arg, { context }) :
                    new hyrest_1.Processed({ value: arg });
                processed[index] = validationResult;
                if (!validationResult.hasErrors) {
                    return;
                }
                errorEncountered = true;
                // Argument can only either be ...
                if (bodyParameters.find(param => param.index === index)) {
                    // ... an injected body.
                    bodyError = validationResult;
                    return;
                }
                const queryParameter = queryParameters.find(param => param.index === index);
                if (queryParameter) {
                    // ... an injected query parameter.
                    queryErrors[queryParameter.name] = validationResult;
                    return;
                }
                // ... or an injected url parameter.
                const urlParameter = urlParameters.find(param => param.index === index);
                paramErrors[urlParameter.name] = validationResult;
            })));
            if (errorEncountered) {
                data = hyrest_1.unprocessableEntity({
                    url: Object.keys(paramErrors).length > 0 ? paramErrors : undefined,
                    body: bodyError,
                    query: Object.keys(queryErrors).length > 0 ? queryErrors : undefined,
                }, "Validation failed.");
            }
            else {
                try {
                    const transformOptions = hyrest_1.getTransforms(route.target, route.property);
                    const processedArgs = processed.map((result, index) => {
                        const bodyParameter = bodyParameters.find(param => param.index === index);
                        const autoPopulate = bodyParameter &&
                            typeof bodyParameter.scope !== "undefined" &&
                            typeof bodyParameter.paramType !== "undefined";
                        if (autoPopulate) {
                            return hyrest_1.populate(bodyParameter.scope, bodyParameter.paramType, result.value);
                        }
                        const parameterTranform = transformOptions.parameterTransforms.get(index);
                        if (parameterTranform) {
                            return parameterTranform(result.value);
                        }
                        return result.value;
                    });
                    const routeResult = yield routeMethod.apply(controllerObject, processedArgs);
                    if (typeof route.scope !== "undefined" && typeof route.returnType !== "undefined") {
                        data = hyrest_1.dump(route.scope, routeResult);
                    }
                    else {
                        data = routeResult;
                    }
                }
                catch (err) {
                    next(err);
                    data = hyrest_1.internalServerError();
                }
            }
            // Respond to the request with the given status code and body.
            const lastCall = hyrest_1.consumeLastCall();
            if (typeof lastCall === "undefined") {
                next(new Error("Route did not return an answer. Make sure the return value is wrapped properly."));
                return;
            }
            const { statusCode, message } = lastCall;
            response.status(statusCode).send({ data, message });
            return;
        });
        // Depending on the HTTP method of the route, make the router listen to the given url and have it
        // call `handler` for each request.
        switch (route.method) {
            case "GET":
                router.get(route.url, handler);
                break;
            case "POST":
                router.post(route.url, handler);
                break;
            case "PATCH":
                router.patch(route.url, handler);
                break;
            case "PUT":
                router.put(route.url, handler);
                break;
            case "DELETE":
                router.delete(route.url, handler);
                break;
            case "HEAD":
                router.head(route.url, handler);
                break;
            case "OPTIONS":
                router.options(route.url, handler);
                break;
            case "TRACE":
                router.trace(route.url, handler);
                break;
            default: throw new Error(`Unknown HTTP method ${route.method}. Take a look at ${route.property}.`);
        }
    });
    router.context = (factory) => {
        contextFactory = factory;
        return router;
    };
    router.authorization = (checker) => {
        authorizationCheck = checker;
        return router;
    };
    router.defaultAuthorizationMode = (defaultMode) => {
        defaultAuthorizationMode = defaultMode;
        return router;
    };
    return router;
}
exports.hyrest = hyrest;
//# sourceMappingURL=middleware.js.map